import contextlib
import functools
import math
import os
import typing
from collections.abc import Callable, Iterable, Iterator, MutableMapping, MutableSet, Sequence, Set
from pathlib import Path

import attrs

if typing.TYPE_CHECKING:
    from puya.options import PuyaOptions

T_A = typing.TypeVar("T_A", bound=attrs.AttrsInstance)


def sha512_256_hash(value: bytes) -> bytes:
    """
    Returns the SHA512/256 hash of a value. This is the hashing algorithm used
    to generate address checksums
    """
    from Cryptodome.Hash import SHA512

    sha = SHA512.new(truncate="256")
    sha.update(value)
    return sha.digest()


def attrs_extend(
    new_type: type[T_A], base_instance: typing.Any, **changes: typing.Any  # noqa: ANN401
) -> T_A:
    """Like attrs.evolve but allows creating a sub-type"""
    base_type = type(base_instance)
    assert issubclass(new_type, base_type)
    old_type_fields = attrs.fields_dict(base_type)
    new_type_fields = attrs.fields(new_type)
    for a in new_type_fields:
        if not a.init:
            continue
        attr_name = a.name  # To deal with private attributes.
        init_name = a.alias
        if init_name not in changes and attr_name in old_type_fields:
            changes[init_name] = getattr(base_instance, attr_name)

    return new_type(**changes)


T = typing.TypeVar("T")


def no_none_in_list(lst: list[T | None]) -> typing.TypeGuard[list[T]]:
    return None not in lst


def no_none_in_tuple(tup: tuple[T | None, ...]) -> typing.TypeGuard[tuple[T, ...]]:
    return None not in tup


@functools.cache
def make_path_relative(*, to: Path, path: str) -> str:
    with contextlib.suppress(ValueError):
        path = str(Path(path).relative_to(to))
    return path.replace(os.sep, "/")


def make_path_relative_to_cwd(path: str | Path) -> str:
    return make_path_relative(to=Path.cwd(), path=str(path))


def unique(items: Iterable[T]) -> list[T]:
    return list(dict.fromkeys(items))


class StableSet(MutableSet[T]):
    __slots__ = ("_data",)

    def __init__(self, *items: T) -> None:
        self._data = dict.fromkeys(items)

    @classmethod
    def from_iter(cls, items: Iterable[T]) -> "StableSet[T]":
        result = StableSet.__new__(StableSet)
        result._data = dict.fromkeys(items)  # noqa: SLF001
        return result

    def __eq__(self, other: object) -> bool:
        if isinstance(other, StableSet):
            return self._data.__eq__(other._data)
        else:
            return self._data.keys() == other

    def __ne__(self, other: object) -> bool:
        if isinstance(other, StableSet):
            return self._data.__ne__(other._data)
        else:
            return self._data.keys() != other

    def __contains__(self, x: object) -> bool:
        return self._data.__contains__(x)

    def __len__(self) -> int:
        return self._data.__len__()

    def __iter__(self) -> Iterator[T]:
        return self._data.__iter__()

    def add(self, value: T) -> None:
        self._data[value] = None

    def discard(self, value: T) -> None:
        self._data.pop(value, None)

    def intersection(self, other: Iterable[T]) -> "StableSet[T]":
        result = StableSet.__new__(StableSet)
        result._data = dict.fromkeys(k for k in self._data if k in other)  # noqa: SLF001
        return result

    def __or__(self, other: Iterable[T]) -> "StableSet[T]":  # type: ignore[override]
        result = StableSet.__new__(StableSet)
        if isinstance(other, StableSet):
            other_data = other._data
        else:
            other_data = dict.fromkeys(other)
        result._data = self._data | other_data
        return result

    def __ior__(self, other: Iterable[T]) -> typing.Self:  # type: ignore[override]
        if isinstance(other, StableSet):
            other_data = other._data
        else:
            other_data = dict.fromkeys(other)
        self._data |= other_data
        return self

    def __sub__(self, other: Set[T]) -> "StableSet[T]":
        result = StableSet.__new__(StableSet)
        if isinstance(other, StableSet):
            data: Iterable[T] = self._data.keys() - other._data.keys()
        else:
            data = (k for k in self._data if k not in other)
        result._data = dict.fromkeys(data)
        return result

    def __repr__(self) -> str:
        return type(self).__name__ + "(" + ", ".join(map(repr, self._data)) + ")"


def determine_out_dir(contract_path: Path, options: "PuyaOptions") -> Path:
    if not options.out_dir:
        out_dir = contract_path
    else:
        # find input path the contract is relative to
        for src_path in options.paths:
            src_path = src_path.resolve()
            src_path = src_path if src_path.is_dir() else src_path.parent
            try:
                relative_path = contract_path.relative_to(src_path)
            except ValueError:
                continue
            # construct a path that maintains a hierarchy to src_path
            out_dir = options.out_dir / relative_path
            if not options.out_dir.is_absolute():
                out_dir = src_path / out_dir
            break
        else:
            # if not relative to any input path
            if options.out_dir.is_absolute():
                out_dir = options.out_dir / contract_path
            else:
                out_dir = contract_path / options.out_dir

    out_dir.mkdir(parents=True, exist_ok=True)
    return out_dir


U = typing.TypeVar("U")


def lazy_setdefault(m: MutableMapping[T, U], /, key: T, default: Callable[[T], U]) -> U:
    """dict.setdefault, but with a callable"""
    try:
        return m[key]
    except KeyError:
        pass
    value = default(key)
    m[key] = value
    return value


_INVERT_ORDERED_BINARY_OP = str.maketrans("<>", "><")


def invert_ordered_binary_op(op: str) -> str:
    return op.translate(_INVERT_ORDERED_BINARY_OP)


def clamp(value: int, *, low: int, high: int) -> int:
    if value < low:
        return low
    if value > high:
        return high
    return value


def bits_to_bytes(bit_size: int) -> int:
    return int(math.ceil(bit_size / 8))


def round_bits_to_nearest_bytes(bit_size: int) -> int:
    return bits_to_bytes(bit_size) * 8


@contextlib.contextmanager
def pushd(new_dir: Path) -> Iterator[None]:
    orig_dir = Path.cwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(orig_dir)


def normalise_path_to_str(path: Path) -> str:
    return str(path).replace("\\", "/")


def biguint_bytes_eval(value: int) -> bytes:
    byte_length = math.ceil(value.bit_length() / 8.0)
    assert byte_length <= 64, "Biguints must be 64 bytes or less"
    big_uint_bytes = value.to_bytes(byteorder="big", length=byte_length)
    return big_uint_bytes


@typing.overload
def coalesce(arg1: T | None, arg2: T, /) -> T: ...


@typing.overload
def coalesce(arg1: T | None, arg2: T | None, arg3: T, /) -> T: ...


@typing.overload
def coalesce(*args: T | None) -> T | None: ...


def coalesce(*args: T | None) -> T | None:
    """Shorthand for `a if a is not None else b`, with eager evaluation as a tradeoff"""
    # REFACTOR: if there's a better way to do the above overloads, we should.
    #           the problem is you can't have a positional argument after *args,
    #           and we want to take the last one's type separately
    for arg in args:
        if arg is not None:
            return arg
    return None


def positive_index(idx: int, seq: Sequence[T]) -> int:
    return idx if idx >= 0 else len(seq) + idx
